﻿using System;
using Microsoft.SPOT;

using GT = Gadgeteer;
using GTM = Gadgeteer.Modules;
using GTI = Gadgeteer.Interfaces;

namespace Gadgeteer.Modules.Gadgeteering
{
    /// <summary>
    /// A ThermalPrinter module for Microsoft .NET Gadgeteer
    /// </summary>
    public class ThermalPrinter : GTM.Module
    {

        private GTI.Serial serial;

        // Note: A constructor summary is auto-generated by the doc builder.
        /// <summary></summary>
        /// <param name="socketNumber">The socket that this module is plugged in to.</param>
        public ThermalPrinter(int socketNumber)
        {
            Socket socket = Socket.GetSocket(socketNumber, true, this, null);
            serial = new GTI.Serial(socket, 19200, GTI.Serial.SerialParity.None, GTI.Serial.SerialStopBits.One, 8, GTI.Serial.HardwareFlowControl.NotRequired, this);
            Initialize();
        }

        /// <summary>
        /// Clears the print buffer, resets parameters to their default values, sets the printer 
        /// to 'standard' mode and deletes any user-defined characters.
        /// </summary>
        public void Initialize()
        {
            if (serial.IsOpen)
                serial.Close();

            serial.Open();

            serial.Write(new byte[] { 0x1B, 0x40 });

            _heatingDotsMax = 64;
            _heatingTime = 800;
            _heatingInterval = 20;

            UpdateHeatingSettings();
        }

        /// <summary>
        /// Prints a message.
        /// </summary>
        /// <param name="message">The message to print.</param>
        public void Print(string message)
        {
            serial.WriteLine(message);
            // Prints the data in the buffer and feeds two new lines.
            serial.Write(0x0A);
        }

        /// <summary>
        /// Prints a Bitmap image.
        /// </summary>
        /// <param name="bitmap"></param>
        public void Print(Bitmap bitmap)
        {
            // 1 pixel is represented by a single bit. Data is sent out in bytes, so the number of bytes 
            // in a row is the pixel width of the bitmap divided by 8. If the pixel width is not a multiple of 8, then
            // add an additional byte (unused pixels will be padded with white).
            int bytesPerRow = (bitmap.Width % 8) == 0 ? (bitmap.Width / 8) : ((bitmap.Width / 8) + 1);
            int numberOfRows = bitmap.Height;

            // Send "Print bitmap with width & height" command
            serial.Write(new byte[] { 0x12, 0x2A, (byte)numberOfRows, (byte)bytesPerRow });

            Debug.Print("Image start");

            for (int y = 0; y < bitmap.Height; y++)
            {
                for (int x = 0; x < bitmap.Width; x += 8)
                {
                    byte pixelByte = 0x00;

                    for (int bit = 0; bit < 8; bit++)
                    {
                        if ((x + bit) < bitmap.Width)
                        {
                            if (bitmap.GetPixel(x + bit, y) != GT.Color.White)
                            {
                                pixelByte |= (byte)(0x01 << (7 - bit));
                            }
                        }
                    }
                    serial.Write(pixelByte);
                }
            }

            serial.Write(0x0A);
        }

        /// <summary>
        ///  Private variables to maintain the state of the printer settings properties.
        /// </summary>
        private int _heatingDotsMax;
        private int _heatingTime;
        private int _heatingInterval;

        /// <summary>
        /// Units are in microseconds (us). Valid range is 30 to 2550. Default is 800us.
        /// Greater heating time results in greater printing density, but also in slower printing speed.
        /// If the heating time is too short, the printing may not be visible. 
        /// </summary>
        public int HeatingTime
        {

            get
            {
                return _heatingTime;
            }

            set
            {
                if (value < 30)
                    _heatingTime = 30;
                else if (value > 2550)
                    _heatingTime = 2550;
                else
                    _heatingTime = value;

                UpdateHeatingSettings();
            }
        }

        /// <summary>
        /// Units are in dots. Valid range is 0 to 2048, in multiples of 8. Default is 64 dots.
        /// The more maximum heating dots, the more peak current the printer will use when printing, but the faster the printing speed. 
        /// </summary>
        public int HeatingDotsMax
        {
            get
            {
                return _heatingDotsMax;
            }

            set
            {
                if (value < 0)
                    _heatingDotsMax = 0;
                else if (value > 2048)
                    _heatingDotsMax = 2048;
                else
                    _heatingDotsMax = value;

                UpdateHeatingSettings();
            }
        }

        /// <summary>
        /// Units are in microseconds (us). Valid range is 0 to 2550. Default is 20us.
        /// Greater heating interval results in clearer printer, but also in slower printing speed.
        /// </summary>
        public int HeatingInterval
        {
            get
            {
                return _heatingInterval;
            }

            set
            {
                if (value < 0)
                    _heatingInterval = 0;
                else if (value > 2550)
                    _heatingInterval = 2550;
                else
                    _heatingInterval = value;

                UpdateHeatingSettings();
            }
        }


        private void UpdateHeatingSettings()
        {
            int heatingDots = (HeatingDotsMax / 8) - 1;
            int heatingTime = HeatingTime / 10;
            int heatingInterval = HeatingInterval / 10;
            serial.Write(new byte[] { 0x1B, 0x37, (byte)heatingDots, (byte)heatingTime, (byte)heatingInterval });
        }

        /// <summary>
        /// Available text styles. Use SetTextStyle method to use these.
        /// </summary>
        [Flags]
        public enum TextStyle : byte
        {
            Regular = 0x00,
            Reverse = 0x02,
            Updown = 0x04,
            Emphasized = 0x08,
            DoubleHeight = 0x10,
            DoubleWidth = 0x20,
            Strikethrough = 0x40
        }

        /// <summary>
        /// Sets the style of any text 
        /// </summary>
        /// <param name="textStyle">You can set multiple text styles by using the "|" (OR) operator.</param>
        public void SetTextStyle(TextStyle textStyle)
        {
            serial.Write(new byte[] { 0x1B, 0x21, (byte)textStyle });
        }

        /// <summary>
        /// Reverse the text printing color.
        /// </summary>
        /// <param name="reverse">White text on black if true, black text on white if false (default).</param>
        public void SetReversePrinting(bool reverse)
        {
            if (reverse)
            {
                serial.Write(new byte[] { 0x1D, 0x42, 1 });
            }
            else
            {
                serial.Write(new byte[] { 0x1D, 0x42, 0 });
            }
        }
    }
}
